import shlex
from enum import Enum
from pathlib import Path

import pytest
from typing_extensions import override

from clypi import Command, Positional, arg
from clypi.configuration import get_config


def parametrize(args: str, cases: list[tuple]):
    def wrapper(fn):
        return pytest.mark.parametrize(args, cases, ids=ids(cases))(fn)

    return wrapper


get_config().help_on_fail = False


class ExampleSub(Command):
    pos2: Positional[tuple[str | Path, ...]]
    flag2: bool = False
    option2: int = 5

    @override
    async def run(self):
        print("subcommand")


class Example(Command):
    pos: Positional[Path]
    flag: bool = arg(default=False, short="f")
    subcommand: ExampleSub | None = None
    option: list[str] = arg(default_factory=list, short="o")

    @override
    async def run(self):
        print("main")


def ids(cases: list[tuple]):
    return list(map(lambda x: shlex.join(x[0]), cases))


COMMAND = [
    (["./some-path"], {"flag": False, "pos": Path("./some-path"), "option": []}),
    (
        ["--flag", "./some-path"],
        {"flag": True, "pos": Path("./some-path"), "option": []},
    ),
    (
        ["./some-path", "--flag"],
        {"flag": True, "pos": Path("./some-path"), "option": []},
    ),
    (
        ["-f", "./some-path"],
        {"flag": True, "pos": Path("./some-path"), "option": []},
    ),
    (
        ["./some-path", "--option", "a"],
        {"flag": False, "pos": Path("./some-path"), "option": ["a"]},
    ),
    (
        ["./some-path", "-o", "a"],
        {"flag": False, "pos": Path("./some-path"), "option": ["a"]},
    ),
    (
        ["./some-path", "--flag", "--option", "a"],
        {"flag": True, "pos": Path("./some-path"), "option": ["a"]},
    ),
    (
        ["./some-path", "--option", "a", "--flag"],
        {"flag": True, "pos": Path("./some-path"), "option": ["a"]},
    ),
    (
        ["./some-path", "--flag", "--option", "a", "b"],
        {"flag": True, "pos": Path("./some-path"), "option": ["a", "b"]},
    ),
    (
        ["./some-path", "--option", "a", "b", "--flag"],
        {"flag": True, "pos": Path("./some-path"), "option": ["a", "b"]},
    ),
    (
        ["./some-path", "-o", "a", "b", "-f"],
        {"flag": True, "pos": Path("./some-path"), "option": ["a", "b"]},
    ),
]


@parametrize("args,expected", COMMAND)
def test_expected_parsing_no_subcommand(args, expected):
    ec = Example.parse(args)
    for k, v in expected.items():
        assert getattr(ec, k) == v


SUBCMD = [
    (
        ["example-sub", "foo"],
        {
            "pos2": ("foo",),
            "flag2": False,
            "option2": 5,
        },
    ),
    (
        ["example-sub", "foo", "bar"],
        {
            "pos2": ("foo", "bar"),
            "flag2": False,
            "option2": 5,
        },
    ),
    (
        ["example-sub", "foo", "bar", "--flag2"],
        {
            "pos2": ("foo", "bar"),
            "flag2": True,
            "option2": 5,
        },
    ),
    (
        ["example-sub", "--flag2", "foo", "bar"],
        {
            "pos2": ("foo", "bar"),
            "flag2": True,
            "option2": 5,
        },
    ),
    (
        ["example-sub", "foo", "bar", "--option2", "6"],
        {
            "pos2": ("foo", "bar"),
            "flag2": False,
            "option2": 6,
        },
    ),
    (
        ["example-sub", "--option2", "6", "foo", "bar"],
        {
            "pos2": ("foo", "bar"),
            "flag2": False,
            "option2": 6,
        },
    ),
    (
        ["example-sub", "--option2", "6", "foo", "bar", "--flag2"],
        {
            "pos2": ("foo", "bar"),
            "flag2": True,
            "option2": 6,
        },
    ),
    (
        ["example-sub", "--flag2", "--option2", "6", "foo", "bar"],
        {
            "pos2": ("foo", "bar"),
            "flag2": True,
            "option2": 6,
        },
    ),
    (
        ["example-sub", "--flag2", "foo", "bar", "--option2", "6"],
        {
            "pos2": ("foo", "bar"),
            "flag2": True,
            "option2": 6,
        },
    ),
]

# Test every COMMAND scenario with every SUBCMD scenario
MERGED = [
    (
        [*cmd_args, *subcmd_args],
        cmd_expected,
        subcmd_expected,
    )
    for subcmd_args, subcmd_expected in SUBCMD
    for cmd_args, cmd_expected in COMMAND
]


@parametrize("args,cmd_expected,subcmd_expected", MERGED)
def test_expected_parsing_subcommand(args, cmd_expected, subcmd_expected):
    ec = Example.parse(args)
    for k, v in cmd_expected.items():
        assert getattr(ec, k) == v

    sc = ec.subcommand
    assert sc is not None
    for k, v in subcmd_expected.items():
        assert getattr(sc, k) == v


@parametrize(
    "args,expected,fails",
    [
        ([], {}, True),
        (["foo"], {}, True),
        (
            ["foo", "--opt", "bar"],
            {"pos": ["foo"], "opt": ["bar"], "opt2": []},
            False,
        ),
        (
            ["foo", "--opt", "bar", "--opt2", "baz"],
            {"pos": ["foo"], "opt": ["bar"], "opt2": ["baz"]},
            False,
        ),
        (
            [
                *(["foo"] * 10),
                "--opt",
                *(["bar"] * 10),
                "--opt2",
                *(["baz"] * 10),
            ],
            {"pos": ["foo"] * 10, "opt": ["bar"] * 10, "opt2": ["baz"] * 10},
            False,
        ),
    ],
)
def test_parse_lists(args, expected, fails):
    class ListCommand(Command):
        pos: Positional[list[str]]
        opt: list[str]
        opt2: list[str] = arg(default_factory=list)

    if fails:
        with pytest.raises(Exception):
            _ = ListCommand.parse(args)
        return

    lc = ListCommand.parse(args)
    assert lc is not None
    for k, v in expected.items():
        lc_v = getattr(lc, k)
        assert lc_v == v
        assert isinstance(lc_v, list)


@parametrize(
    "args,expected,fails",
    [
        ([], {}, True),
        (["foo"], {}, True),
        (["foo", "--opt", "bar"], {}, True),
        (
            ["foo", "--opt", "bar", "baz"],
            {"pos": ("foo",), "opt": ("bar", "baz"), "opt2": tuple()},
            False,
        ),
        (
            ["foo", "--opt", "bar", "baz", "--opt2", "qux"],
            {"pos": ("foo",), "opt": ("bar", "baz"), "opt2": ("qux",)},
            False,
        ),
        ([*(["foo"] * 2), "--opt", "bar", "--opt2", "qux"], {}, True),
        (["foo", "--opt", *(["bar"] * 3), "--opt2", "qux"], {}, True),
        (["foo", "--opt", *(["bar"] * 2), "--opt2", *(["qux"] * 10)], {}, False),
    ],
)
def test_parse_tuples(args, expected, fails):
    class TupleCommand(Command):
        pos: Positional[tuple[str]]
        opt: tuple[str, str]
        opt2: tuple[str, ...] = arg(default_factory=tuple)

    if fails:
        with pytest.raises(Exception):
            _ = TupleCommand.parse(args)
        return

    lc = TupleCommand.parse(args)
    assert lc is not None
    for k, v in expected.items():
        lc_v = getattr(lc, k)
        assert lc_v == v
        assert isinstance(lc_v, tuple)


class Env(Enum):
    QA = 1
    PROD = 2


@parametrize(
    "args,expected,fails",
    [
        ([], {}, True),
        (["foo"], {}, True),
        (["qa"], {}, True),
        (["qa", "--env2", "bar"], {}, True),
        (["qa", "--env2", "prod", "--env3", "foo"], {}, True),
        (
            ["qa", "--env2", "prod"],
            {"env": Env.QA, "env2": Env.PROD, "env3": Env.PROD},
            False,
        ),
        (
            ["prod", "--env2", "qa", "--env3", "prod"],
            {"env": Env.PROD, "env2": Env.QA, "env3": Env.PROD},
            False,
        ),
    ],
)
def test_parse_enums(args, expected, fails):
    class EnumCommand(Command):
        env: Positional[Env]
        env2: Env
        env3: Env = Env.PROD

    if fails:
        with pytest.raises(Exception):
            _ = EnumCommand.parse(args)
        return

    ec = EnumCommand.parse(args)
    assert ec is not None
    for k, v in expected.items():
        lc_v = getattr(ec, k)
        assert lc_v == v
        assert isinstance(lc_v, Env)



import asyncio
import re
from pathlib import Path

import clypi
from clypi import ClypiException, Command, Positional, Spinner, arg


async def from_requirements(file: Path):
    packages_with_versions: dict[str, str] = {}
    for line in file.read_text().split():
        package = re.search(r"(\w+)[>=<]+([0-9\.]+)", line)
        if not package:
            continue
        packages_with_versions[package.group(1)] = package.group(2)

    await _install_packages(packages_with_versions)


async def from_packages(packages: list[str]):
    packages_with_versions: dict[str, str] = {}

    clypi.print("\nAdded new packages", fg="blue", bold=True)
    for p in packages:
        package = re.search(r"(\w+)[>=<]+([0-9\.]+)", p)
        if not package:
            continue
        packages_with_versions[package.group(1)] = package.group(2)

    await _install_packages(packages_with_versions)


async def _install_packages(packages: dict[str, str]):
    async with Spinner("Installing packages", capture=True):
        for name, version in packages.items():
            print("Installed", name)
            await asyncio.sleep(0.3)

    clypi.print("\nAdded new packages", fg="blue", bold=True)
    for name, version in packages.items():
        icon = clypi.style("+", fg="green", bold=True)
        print(f"[{icon}] {name} {version}")


class Add(Command):
    """Add dependencies to the project"""

    packages: Positional[list[str]] = arg(
        default_factory=list,
        help="The packages to add, as PEP 508 requirements (e.g., `ruff==0.5.0`)",
    )
    requirements: Path | None = arg(
        default=None,
        short="r",
        help="Add all packages listed in the given `requirements.txt` files",
    )
    dev: bool = arg(
        default=False, help="Add the requirements to the development dependency group"
    )

    async def run(self) -> None:
        clypi.print("Running `uv add` command...\n", fg="blue", bold=True)

        # Download from requirements.txt file
        if self.requirements:
            await from_requirements(self.requirements)

        # Download positional args
        elif self.packages:
            await from_packages(self.packages)

        else:
            raise ClypiException("One of requirements or packages is required!")



import typing as t

from clypi._data.boxes import Boxes as _Boxes
from clypi.colors import remove_style

Boxes = _Boxes


def _real_len(s: str) -> int:
    s = remove_style(s)
    return len(s)


def _ljust(s: str, width: int):
    len = _real_len(s)
    diff = max(0, width - len)
    return s + " " * diff


def _rjust(s: str, width: int):
    len = _real_len(s)
    diff = max(0, width - len)
    return " " * diff + s


def _center(s: str, width: int):
    len = _real_len(s)
    diff = max(0, width - len)
    right = diff // 2
    left = diff - right
    return " " * left + s + " " * right


AlignType: t.TypeAlias = t.Literal["left", "center", "right"]


def align(s: str, alignment: AlignType, width: int) -> str:
    if alignment == "left":
        return _ljust(s, width)
    if alignment == "right":
        return _rjust(s, width)
    return _center(s, width)



from __future__ import annotations

import typing as t
from dataclasses import dataclass
from functools import cached_property

from clypi import boxed, indented, stack
from clypi._cli import type_util
from clypi.exceptions import format_traceback

if t.TYPE_CHECKING:
    from clypi.cli import Command, _Argument


@dataclass
class ProgramConfig:
    prog: str


def _ext(ls: list[str], s: str | list[str] | None) -> None:
    if isinstance(s, str):
        ls.append(s)
    elif isinstance(s, list):
        ls.extend(s)
    return None


class Formatter(t.Protocol):
    def format_help(
        self,
        prog: list[str],
        description: str | None,
        epilog: str | None,
        options: list[_Argument],
        positionals: list[_Argument],
        subcommands: list[type[Command]],
        exception: Exception | None,
    ) -> str: ...


@dataclass
class ClypiFormatter:
    def __init__(self, boxed: bool = True) -> None:
        self.boxed = boxed

    @cached_property
    def theme(self):
        from clypi.configuration import get_config

        return get_config().theme

    def _format_option(self, option: _Argument) -> tuple[str, ...]:
        usage = self.theme.long_option(option.display_name)
        short_usage = self.theme.short_option(
            option.short_display_name if option.short else ""
        )
        type_str = self.theme.type_str(type_util.type_to_str(option.arg_type).upper())
        help = option.help or ""

        return usage, short_usage, type_str, help

    def _format_options(self, options: list[_Argument]) -> list[str] | None:
        if not options:
            return None

        usage: list[str] = []
        short_usage: list[str] = []
        type_str: list[str] = []
        help: list[str] = []
        for o in options:
            u, su, ts, hp = self._format_option(o)
            usage.append(u)
            short_usage.append(su)
            type_str.append(ts)
            help.append(hp)

        stacked_options = stack(usage, short_usage, type_str, help, lines=True)
        if not self.boxed:
            section_title = self.theme.section_title("Options")
            return [section_title] + indented(stacked_options) + [""]

        return list(boxed(stacked_options, title="Options"))

    def _format_positional(self, positional: _Argument) -> t.Any:
        name = self.theme.positional(positional.name)
        help = positional.help or ""
        type_str = self.theme.type_str(
            type_util.type_to_str(positional.arg_type).upper()
        )

        return name, type_str, help

    def _format_positionals(
        self, positionals: list[_Argument]
    ) -> list[str] | str | None:
        if not positionals:
            return None

        name: list[str] = []
        type_str: list[str] = []
        help: list[str] = []
        for p in positionals:
            n, ts, hp = self._format_positional(p)
            name.append(n)
            type_str.append(ts)
            help.append(hp)

        stacked_positionals = stack(name, type_str, help, lines=True)
        if not self.boxed:
            section_title = self.theme.section_title("Arguments")
            return [section_title] + indented(stacked_positionals) + [""]

        return list(boxed(stacked_positionals, title="Arguments"))

    def _format_subcommand(self, subcmd: type[Command]) -> tuple[str, str]:
        name = self.theme.subcommand(subcmd.prog())
        help = subcmd.help() or ""
        return name, help

    def _format_subcommands(
        self, subcommands: list[type[Command]]
    ) -> list[str] | str | None:
        if not subcommands:
            return None

        name: list[str] = []
        help: list[str] = []
        for p in subcommands:
            n, hp = self._format_subcommand(p)
            name.append(n)
            help.append(hp)

        stacked_subcommands = stack(name, help, lines=True)
        if not self.boxed:
            section_title = self.theme.section_title("Subcommands")
            return [section_title] + indented(stacked_subcommands) + [""]

        return list(boxed(stacked_subcommands, title="Subcommands"))

    def _format_header(
        self,
        prog: list[str],
        options: list[_Argument],
        positionals: list[_Argument],
        subcommands: list[type[Command]],
    ) -> list[str] | str | None:
        prefix = self.theme.usage("Usage:")
        prog_str = self.theme.prog(" ".join(prog))

        option = " [" + self.theme.long_option("OPTIONS") + "]" if options else ""
        command = self.theme.subcommand(" COMMAND") if subcommands else ""
        positional = (
            " " + " ".join(self.theme.positional(p.name.upper()) for p in positionals)
            if positionals
            else ""
        )

        return [f"{prefix} {prog_str}{option}{command}{positional}"]

    def _format_description(self, description: str | None) -> list[str] | str | None:
        if not description:
            return None
        return [description, ""]

    def _format_exception(self, exception: Exception | None) -> list[str] | str | None:
        if not exception:
            return None

        if not self.boxed:
            section_title = self.theme.section_title("Error")
            return [section_title] + indented(format_traceback(exception)) + [""]
        return list(boxed(format_traceback(exception), title="Error", color="red"))

    def format_help(
        self,
        prog: list[str],
        description: str | None,
        epilog: str | None,
        options: list[_Argument],
        positionals: list[_Argument],
        subcommands: list[type[Command]],
        exception: Exception | None,
    ) -> str:
        lines: list[str] = []

        # Header
        _ext(lines, self._format_header(prog, options, positionals, subcommands))
        _ext(lines, "")

        # Description
        _ext(lines, self._format_description(description))

        # Subcommands
        _ext(lines, self._format_subcommands(subcommands))

        # Positionals
        _ext(lines, self._format_positionals(positionals))

        # Options
        _ext(lines, self._format_options(options))

        # Epilog
        _ext(lines, self._format_description(epilog))

        # Exceptions
        _ext(lines, self._format_exception(exception))

        return "\n".join(lines)



import typing as t
from pathlib import Path

from clypi import Command, arg


class MySub(Command):
    foo: int = 1


class MyCommand(Command):
    subcommand: MySub
    foo: bool
    bar: str = arg(help="help!")
    baz: Path = arg(default=Path.cwd())
    qux: list[str] = arg(default_factory=list)


cli = MyCommand.parse()

t.assert_type(cli, MyCommand)
t.assert_type(cli.foo, bool)
t.assert_type(cli.bar, str)
t.assert_type(cli.baz, Path)
t.assert_type(cli.qux, list[str])

t.assert_type(cli.subcommand, MySub)
t.assert_type(cli.subcommand.foo, int)

t.assert_type(arg(default=None), None)
t.assert_type(arg(default=5), int)
t.assert_type(arg(default_factory=int), int)


def parser(x: t.Any) -> int | list[str]:
    if x == 1:
        return x
    return ["a"]


t.assert_type(arg(parser=parser), int | list[str])



from clypi import Command, Positional, arg


class Lint(Command):
    files: Positional[tuple[str, ...]]
    verbose: bool = arg(...)  # Comes from MyCli but I want to use it too

    async def run(self):
        print(f"Linting {', '.join(self.files)} and {self.verbose=}")


class MyCli(Command):
    """
    my-cli is a very nifty demo CLI tool
    """

    subcommand: Lint | None = None
    verbose: bool = arg(
        help="Wether to show extra logs",
        prompt="Do you want to see extra logs?",
        default=False,
        short="v",  # User can pass in --verbose or -v
    )

    async def run(self):
        print(f"Running the main command with {self.verbose}")


if __name__ == "__main__":
    cli: MyCli = MyCli.parse()
    cli.start()



import clypi
from clypi import Command


class Install(Command):
    """Install packages into an environment"""

    async def run(self) -> None:
        clypi.print("Running `uv pip install` command...", fg="blue")


class Uninstall(Command):
    """Uninstall packages from an environment"""

    async def run(self) -> None:
        clypi.print("Running `uv pip uninstall` command...", fg="blue")


class Freeze(Command):
    """List, in requirements format, packages installed in an environment"""

    async def run(self) -> None:
        clypi.print("Running `uv pip freeze` command...", fg="blue")


class List(Command):
    """List, in tabular format, packages installed in an environment"""

    async def run(self) -> None:
        clypi.print("Running `uv pip list` command...", fg="blue")


class Pip(Command):
    """Manage Python packages with a pip-compatible interface"""

    subcommand: Install | Uninstall | Freeze | List



from __future__ import annotations

import os
import shlex
import sys
import typing as t
from abc import ABC, abstractmethod
from pathlib import Path
from textwrap import dedent

import clypi
from clypi._cli.parser import parse_as_attr

if t.TYPE_CHECKING:
    from clypi.cli import Command

_CLYPI_CURRENT_ARGS = "_CLYPI_CURRENT_ARGS"


class AutocompleteInstaller(ABC):
    """
    The basic idea for autocomplete with clypi is that we'll configure
    the shell to call the users' CLI with whatever the user has typed so far. Then,
    we will try to parse as much as we can from what the user gave us, and when we can't
    anymore we just list the options the user has. At that point, we will have traversed
    enough into the right subcommand the user wants to autocomplete
    """

    def __init__(self, command: type[Command]) -> None:
        self.prog = command.prog()
        self._options = list(command.options().values())
        self._subcommands = list(command.subcommands().values())

    @abstractmethod
    def path(self) -> Path: ...

    @abstractmethod
    def script(self) -> str: ...

    def list_arguments(self) -> None:
        sys.stdout.write(
            "\n".join(
                [
                    *[s.prog() for s in self._subcommands if s],
                    *[p.display_name for p in self._options],
                ]
            ),
        )
        sys.stdout.flush()
        sys.exit(0)

    @property
    def gen_args(self) -> str:
        # We use get_current_args to pass in what the user has typed so far
        get_current_args = f"{_CLYPI_CURRENT_ARGS}=(commandline -cp)"
        return f"env {get_current_args} {self.prog}"

    def install(self) -> None:
        p = self.path()
        p.parent.mkdir(parents=True, exist_ok=True)
        with open(p, "w") as f:
            f.write(self.script())
        self.post_install(p)
        clypi.print(
            "Successfully installed autocomplete for fish", fg="green", bold=True
        )
        clypi.print(f"  󰘍 {self.path()}")
        sys.exit(0)

    def post_install(self, path: Path):
        return None


class FishInstaller(AutocompleteInstaller):
    def path(self) -> Path:
        return Path.home() / ".config" / "fish" / "completions" / f"{self.prog}.fish"

    def script(self) -> str:
        return f'complete -c {self.prog} --no-files -a "({self.gen_args})" -n "{self.gen_args}"'


class BashInstaller(AutocompleteInstaller):
    def path(self) -> Path:
        base = Path("/etc/bash_completion.d/")
        if Path("/usr/local/etc/bash_completion.d").exists():
            base = Path("/usr/local/etc/bash_completion.d")
        return base / self.prog

    def post_install(self, path: Path):
        bashrc = Path.home() / ".bashrc"
        with open(bashrc, "a+") as f:
            for line in f.readline():
                if str(path) in line:
                    return

            f.write(f"source '{path}'")

    def script(self) -> str:
        return dedent(
            """
            _complete_%(prog)s() {
                _script_commands=$(env %(env_var)s="${COMP_WORDS[*]}" $1)
                local cur="${COMP_WORDS[COMP_CWORD]}"
                COMPREPLY=( $(compgen -W "${_script_commands}" -- ${cur}) )
            }

            complete -o default -F _complete_%(prog)s %(prog)s
            """
            % dict(prog=self.prog, env_var=_CLYPI_CURRENT_ARGS)
        ).strip()


class ZshInstaller(AutocompleteInstaller):
    def path(self) -> Path:
        return Path.home() / ".zfunc" / f"_{self.prog}"

    def post_install(self, path: Path):
        autoload_comp = "fpath+=~/.zfunc; autoload -Uz compinit; compinit"
        zshrc = Path.home() / ".zshrc"
        with open(zshrc, "a+") as f:
            for line in f.readline():
                if autoload_comp in line:
                    return

            f.write(autoload_comp)

    def script(self) -> str:
        return dedent(
            """
            #compdef %(prog)s

            _complete_%(prog)s() {
              IFS=$'\\n' completions=( $(env %(env_var)s="${words[1,$CURRENT]}" %(prog)s) )

              local -a filtered
              for item in "${completions[@]}"; do
                if [[ $item == ${words[$CURRENT]}* ]]; then
                  filtered+=("$item")
                fi
              done
              compadd -U -V unsorted -a filtered
            }

            compdef _complete_%(prog)s %(prog)s
            """
            % dict(prog=self.prog, env_var=_CLYPI_CURRENT_ARGS)
        ).strip()


def get_installer(command) -> AutocompleteInstaller:
    shell = Path(os.environ["SHELL"]).name
    if shell == "fish":
        return FishInstaller(command)
    if shell == "bash":
        return BashInstaller(command)
    if shell == "zsh":
        return ZshInstaller(command)
    raise ValueError(f"Autocomplete is not supported for shell '{shell}'")


def get_autocomplete_args() -> list[str] | None:
    if args := os.environ.get(_CLYPI_CURRENT_ARGS):
        return shlex.split(args)[1:]
    return None


def list_arguments(command: type[Command]):
    get_installer(command).list_arguments()


def requested_autocomplete_install(args: t.Sequence[str]) -> bool:
    if not args:
        return False
    parsed = parse_as_attr(args[-1])
    return parsed.is_long_opt() and parsed.value == "install_autocomplete"


def install_autocomplete(command: type[Command]):
    get_installer(command).install()



import asyncio
import io
import sys
import typing as t
from contextlib import AbstractAsyncContextManager, AbstractContextManager, suppress

from typing_extensions import override

import clypi
from clypi._data.spinners import Spin as _Spin
from clypi.colors import ColorType
from clypi.const import ESC

MOVE_START = f"{ESC}1G"
DEL_LINE = f"{ESC}0K"

Spin = _Spin


class _PerLineIO(io.TextIOBase):
    def __init__(self, *args, new_line_cb: t.Callable[[str], None], **kwargs) -> None:
        """
        A string buffer that captures text and calls the callback `new_line_cb`
        on every line written to the buffer. Useful to redirect stdout and stderr
        but only print them nicely on every new line.
        """
        super().__init__(*args, **kwargs)
        self._new_line_cb = new_line_cb
        self.buffer: list[str] = []

    @override
    def write(self, s: str, /) -> int:
        """
        When we get a string, split it by new lines, submit every line we've
        collected and keep the remainder for future writes
        """
        parts = s.split("\n")

        # If there's a buffer, there's a half-way sentence there, so we merge it
        if self.buffer:
            self.buffer[0] += parts[0]
            self.buffer.extend(parts[1:])
        else:
            self.buffer = parts

        if len(self.buffer) > 1:
            for i in range(0, len(self.buffer) - 1):
                if line := self.buffer[i]:
                    self._new_line_cb(line)
            self.buffer = self.buffer[-1:]
        return 0

    @override
    def flush(self) -> None:
        """
        If flush is called, print whatever we have even if there's no new line
        """
        if self.buffer[0]:
            self._new_line_cb(self.buffer[0])
        self.buffer = []


class RedirectStdPipe(AbstractContextManager):
    def __init__(
        self,
        pipe: t.Literal["stdout", "stderr"],
        target: t.Callable[[str], t.Any],
    ) -> None:
        """
        Given a pipe (stdout or stderr) and a callback function, it redirects
        each line from the pipe into the callback. Useful to redirect users'
        outputs to a custom function without them needing to directly call it.
        """
        self._pipe = pipe
        self._original = getattr(sys, pipe)
        self._new = _PerLineIO(new_line_cb=target)

    @override
    def __enter__(self) -> None:
        self.start()

    def start(self) -> None:
        setattr(sys, self._pipe, self._new)

    @override
    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def stop(self) -> None:
        setattr(sys, self._pipe, self._original)

    def write(self, s: str):
        self._original.write(s)

    def flush(self):
        self._original.flush()


@t.final
class Spinner(AbstractAsyncContextManager):
    def __init__(
        self,
        title: str,
        animation: Spin | list[str] = Spin.DOTS,
        prefix: str = " ",
        suffix: str = "…",
        speed: float = 1,
        capture: bool = False,
    ) -> None:
        """
        A context manager that lets you run async code while nicely
        displaying a spinning animation. Using `capture=True` will
        capture all the stdout and stderr written during the spinner
        and display it nicely.
        """

        self.animation = animation
        self.prefix = prefix
        self.suffix = suffix
        self.title = title

        self._task: asyncio.Task[None] | None = None
        self._manual_exit: bool = False
        self._frame_idx: int = 0
        self._refresh_rate = 0.7 / speed / len(self._frames)

        # For capturing stdout, stderr
        self._capture = capture
        self._stdout = RedirectStdPipe("stdout", self.log)
        self._stderr = RedirectStdPipe("stderr", self.log)

    async def __aenter__(self):
        if self._capture:
            self._stdout.start()
            self._stderr.start()

        self._task = asyncio.create_task(self._spin())
        return self

    @override
    async def __aexit__(self, _type, value, traceback):
        # If a user already called `.done()`, leaving the closure
        # should not re-trigger a re-render
        if self._manual_exit:
            return

        if any([_type, value, traceback]):
            await self.fail()
        else:
            await self.done()

    def _print(
        self,
        msg: str,
        icon: str | None = None,
        color: ColorType | None = None,
        end: str = "",
    ):
        # Build the line being printed
        icon = clypi.style(icon + " ", fg=color) if icon else ""
        msg = f"{self.prefix}{icon}{msg}{end}"

        # Wipe the line for next render
        self._stdout.write(MOVE_START)
        self._stdout.write(DEL_LINE)

        # Write msg and flush
        self._stdout.write(msg)
        self._stdout.flush()

    def _render_frame(self):
        self._print(
            self.title + self.suffix,
            icon=self._frames[self._frame_idx],
            color="blue",
        )

    @property
    def _frames(self) -> list[str]:
        return (
            self.animation.value if isinstance(self.animation, Spin) else self.animation
        )

    async def _spin(self) -> None:
        while True:
            self._frame_idx = (self._frame_idx + 1) % len(self._frames)
            self._render_frame()
            await asyncio.sleep(self._refresh_rate)

    async def _exit(self, msg: str | None = None, success: bool = True):
        if t := self._task:
            t.cancel()
            with suppress(asyncio.CancelledError):
                await t

        # Stop capturing stdout/stderrr
        if self._capture:
            self._stdout.stop()
            self._stderr.stop()

        color: ColorType = "green" if success else "red"
        icon = "✔️" if success else "×"
        self._print(msg or self.title, icon=icon, color=color, end="\n")

    async def done(self, msg: str | None = None):
        self._manual_exit = True
        await self._exit(msg)

    async def fail(self, msg: str | None = None):
        self._manual_exit = True
        await self._exit(msg, success=False)

    def log(
        self,
        msg: str,
        icon: str = "   ┃",
        color: ColorType | None = None,
        end="\n",
    ):
        """
        Log a message nicely from inside a spinner. If `capture=True`, you can
        simply use `print("foo")`.
        """
        self._print(msg.rstrip(), icon=icon, color=color, end=end)
        self._render_frame()

    async def pipe(
        self,
        pipe: asyncio.StreamReader | None,
        color: ColorType = "blue",
        prefix: str = "",
    ) -> None:
        """
        Pass in an async pipe for the spinner to display
        """
        if not pipe:
            return

        while True:
            line = await pipe.readline()
            if not line:
                break

            msg = f"{prefix} {line.decode()}" if prefix else line.decode()
            self.log(msg, color=color)


P = t.ParamSpec("P")


def spinner(
    title: str,
    animation: Spin | list[str] = Spin.DOTS,
    prefix: str = " ",
    suffix: str = "…",
    speed: float = 1,
    capture: bool = False,
) -> t.Callable[[t.Callable[P, t.Awaitable]], t.Callable[P, t.Awaitable]]:
    """
    Utility decorator to wrap a function and display a Spinner while it's running.
    """

    def wrapper(fn: t.Callable[P, t.Awaitable]) -> t.Callable[P, t.Awaitable]:
        async def inner(
            *args: P.args, **kwargs: P.kwargs
        ) -> t.Callable[P, t.Awaitable]:
            async with Spinner(
                title=title,
                animation=animation,
                prefix=prefix,
                suffix=suffix,
                speed=speed,
                capture=capture,
            ):
                return await fn(*args, **kwargs)

        return inner

    return wrapper



from pathlib import Path

from typing_extensions import override

from clypi import Command, Positional, arg


class ExampleSubCommand(Command):
    """Some sample docs"""

    positional: Positional[tuple[str | Path, ...]]

    async def run(self):
        print("subcommand")


class ExampleCommand(Command):
    """
    Some sample documentation for the main command
    """

    flag: bool = False
    subcommand: ExampleSubCommand | None = None
    option: list[str] = arg(help="A list of strings please", default_factory=list)

    @override
    @classmethod
    def prog(cls):
        return "example"

    @override
    @classmethod
    def epilog(cls):
        return "Some text to display after..."

    async def run(self):
        print("main")


def test_expected_base():
    assert ExampleCommand.help() == "Some sample documentation for the main command"
    assert ExampleCommand.prog() == "example"
    assert ExampleCommand.epilog() == "Some text to display after..."


def test_expected_options():
    opts = ExampleCommand.options()
    assert len(opts) == 2

    assert opts["flag"].name == "flag"
    assert opts["flag"].arg_type is bool
    assert opts["flag"].nargs == 0

    assert opts["option"].name == "option"
    assert opts["option"].arg_type == list[str]
    assert opts["option"].nargs == "*"


def test_expected_positional():
    pos = ExampleSubCommand.positionals()
    assert len(pos) == 1

    assert pos["positional"].name == "positional"
    assert pos["positional"].arg_type == Positional[tuple[str | Path, ...]]
    assert pos["positional"].nargs == "+"


def test_expected_subcommands():
    pos = ExampleCommand.subcommands()
    assert len(pos) == 2

    assert pos[None] is None

    sub = pos["example-sub-command"]
    assert sub is ExampleSubCommand
    assert sub.prog() == "example-sub-command"
    assert sub.help() == "Some sample docs"


def test_expected_cls_introspection():
    assert ExampleCommand.option == []



import sys

import clypi
from clypi import ClypiConfig, ClypiFormatter, Command, Styler, Theme, arg, configure
from examples.uv.add import Add
from examples.uv.init import Init
from examples.uv.pip import Pip
from examples.uv.remove import Remove


class Uv(Command):
    """
    A clone of an extremely fast Python package manager.
    """

    subcommand: Add | Init | Pip | Remove | None
    quiet: bool = arg(default=False, short="q", help="Do not print any output")
    version: bool = arg(default=False, short="V", help="Display the uv version")

    async def run(self) -> None:
        if self.version:
            clypi.print("clypi's UV 0.0.1", fg="green")
            sys.exit(0)

        if not self.quiet:
            self.print_help()


if __name__ == "__main__":
    configure(
        ClypiConfig(
            theme=Theme(
                usage=Styler(fg="green", bold=True),
                prog=Styler(fg="cyan", bold=True),
                section_title=Styler(fg="green", bold=True),
                subcommand=Styler(fg="cyan"),
                long_option=Styler(fg="cyan"),
                short_option=Styler(fg="cyan"),
                positional=Styler(fg="cyan"),
                type_str=Styler(fg="cyan"),
                prompts=Styler(fg="green", bold=True),
            ),
            help_formatter=ClypiFormatter(boxed=False),
        )
    )

    uv = Uv.parse()
    uv.start()



import typing as t
from importlib import import_module

if t.TYPE_CHECKING:
    from clypi.align import AlignType, align
    from clypi.boxed import Boxes, boxed
    from clypi.cli import ClypiFormatter, Command, Formatter, Positional, arg
    from clypi.colors import ALL_COLORS, Styler, print, style
    from clypi.configuration import ClypiConfig, Theme, configure, get_config
    from clypi.exceptions import (
        AbortException,
        ClypiException,
        MaxAttemptsException,
        format_traceback,
        print_traceback,
    )
    from clypi.indented import indented
    from clypi.prompts import (
        Parser,
        confirm,
        prompt,
    )
    from clypi.spinner import Spinner
    from clypi.stack import stack

__all__ = (
    "ALL_COLORS",
    "AbortException",
    "AlignType",
    "Boxes",
    "ClypiConfig",
    "ClypiException",
    "ClypiFormatter",
    "Command",
    "Formatter",
    "MaxAttemptsException",
    "Parser",
    "Positional",
    "Spinner",
    "Styler",
    "Theme",
    "align",
    "arg",
    "boxed",
    "configure",
    "confirm",
    "format_traceback",
    "get_config",
    "indented",
    "print",
    "print_traceback",
    "prompt",
    "stack",
    "style",
)

_dynamic_imports: dict[str, tuple[str | None, str]] = {
    "ALL_COLORS": (__spec__.parent, ".colors"),
    "AbortException": (__spec__.parent, ".exceptions"),
    "AlignType": (__spec__.parent, ".align"),
    "Boxes": (__spec__.parent, ".boxed"),
    "ClypiConfig": (__spec__.parent, ".configuration"),
    "ClypiException": (__spec__.parent, ".exceptions"),
    "ClypiFormatter": (__spec__.parent, ".cli"),
    "Command": (__spec__.parent, ".cli"),
    "Formatter": (__spec__.parent, ".cli"),
    "MaxAttemptsException": (__spec__.parent, ".exceptions"),
    "Parser": (__spec__.parent, ".prompts"),
    "Positional": (__spec__.parent, ".cli"),
    "Spinner": (__spec__.parent, ".spinner"),
    "Styler": (__spec__.parent, ".colors"),
    "Theme": (__spec__.parent, ".configuration"),
    "align": (__spec__.parent, ".align"),
    "arg": (__spec__.parent, ".cli"),
    "boxed": (__spec__.parent, ".boxed"),
    "configure": (__spec__.parent, ".configuration"),
    "confirm": (__spec__.parent, ".prompts"),
    "format_traceback": (__spec__.parent, ".exceptions"),
    "get_config": (__spec__.parent, ".configuration"),
    "indented": (__spec__.parent, ".indented"),
    "print": (__spec__.parent, ".colors"),
    "print_traceback": (__spec__.parent, ".exceptions"),
    "prompt": (__spec__.parent, ".prompts"),
    "stack": (__spec__.parent, ".stack"),
    "style": (__spec__.parent, ".colors"),
}


def __getattr__(attr: str) -> object:
    """
    Dynamically re-exports modules with lazy loading
    """

    dynamic_attr = _dynamic_imports.get(attr)
    if dynamic_attr is None:
        raise AttributeError(f"module {__name__!r} has no attribute {attr!r}")

    package, module_name = dynamic_attr

    # Import module and find the object
    module = import_module(module_name, package=package)
    result = getattr(module, attr)

    # Import everything else in that module
    g = globals()
    for k, (_, other) in _dynamic_imports.items():
        if other == module_name:
            g[k] = getattr(module, k)
    return result


def __dir__() -> list[str]:
    return list(__all__)



from dataclasses import dataclass
from enum import Enum


@dataclass
class Box:
    """
    tl  x myt x  tr
    y             y
    mxl x mm  x mxr
    y             y
    bl  x myb x  br
    """

    tl: str
    tr: str
    bl: str
    br: str
    x: str
    y: str


_ROUNDED = Box(
    tl="╭",
    tr="╮",
    bl="╰",
    br="╯",
    x="─",
    y="│",
)

_THIN = Box(
    tl="┌",
    tr="┐",
    bl="└",
    br="┘",
    x="─",
    y="│",
)

_HEAVY = Box(
    tl="┏",
    tr="┓",
    bl="┗",
    br="┛",
    x="━",
    y="┃",
)


class Boxes(Enum):
    ROUNDED = _ROUNDED
    THIN = _THIN
    HEAVY = _HEAVY

    def human_name(self):
        name = self.name
        return " ".join(p.capitalize() for p in name.split("_"))



from __future__ import annotations

import io
import re
import sys
from contextlib import contextmanager
from datetime import date, datetime, timedelta

import pytest
from pytest import mark

import clypi
from clypi import AbortException, MaxAttemptsException, Parser
from clypi.configuration import get_config


@contextmanager
def replace_stdin(answers: list[str] | str = []):
    answers = [answers] if isinstance(answers, str) else answers
    target = "\n".join(answers)

    orig = sys.stdin
    sys.stdin = io.StringIO(target)
    yield
    sys.stdin = orig


@contextmanager
def replace_stdout():
    orig = sys.stdout
    new = io.StringIO()
    sys.stdout = new
    yield new
    sys.stdout = orig


def _escape_ansi(line: str) -> str:
    """
    https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
    """
    ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", line)


def assert_prompted_times(prompted: io.StringIO, times: int):
    __tracebackhide__ = True
    text = _escape_ansi(prompted.getvalue())
    lines = list(filter(None, text.split(": ")))
    assert len(lines) == times


def _raise_error(x: str) -> None:
    raise ValueError(f"Invalid number {x}")


class TestCase:
    def setup_method(self):
        conf = get_config()
        conf.help_on_fail = False
        conf.nice_errors = tuple()

    @mark.parametrize(
        "expected",
        [True, False],
    )
    def test_confirm_with_default(self, expected: bool):
        with replace_stdin("\n") as _:
            assert clypi.confirm("What's your name?", default=expected) == expected

    @mark.parametrize(
        "answer, expected",
        [
            ("y", True),
            ("Yes", True),
            ("true", True),
            ("N", False),
            ("no", False),
            ("False", False),
        ],
    )
    def test_confirm(self, answer: str, expected: bool):
        with replace_stdin(answer) as _:
            assert clypi.confirm("What's your name?") == expected

    def test_confirm_aborts(self):
        with replace_stdin("n") as _:
            with pytest.raises(AbortException):
                _ = clypi.confirm("What's your name?", abort=True)

    @mark.parametrize(
        "answer,expected",
        [
            ("Alice\n", "Alice"),
            ("\n", "John Doe"),
        ],
    )
    def test_prompt_with_default(self, answer: str, expected: str):
        with replace_stdin(answer) as _:
            assert clypi.prompt("What's your name?", default="John Doe") == expected

    @mark.parametrize(
        "answers,expected,times",
        [
            (["Alice"], "Alice", 1),
            (["", "", "Alice"], "Alice", 3),
        ],
    )
    def test_prompt_with_no_default(
        self, answers: list[str], expected: str, times: int
    ):
        with replace_stdin(answers) as _, replace_stdout() as stdout:
            assert clypi.prompt("What's your name?") == expected
            assert_prompted_times(stdout, times)

    def test_prompt_with_parser(self):
        with replace_stdin("42") as _:
            res = clypi.prompt("Some prompt", parser=int)
            assert isinstance(res, int)

    @mark.parametrize(
        "answer,parser",
        [
            ("Alice", int),
            ("42 days", int),
            ("42 asd", timedelta),
            ("42", timedelta),
            ("202-01-01", date),
            ("2021-01-01T00:00", datetime),
        ],
        ids=[
            "Str as Int",
            "TimeDelta as Int",
            "Invalid TimeDelta",
            "Int as TimeDelta",
            "Invalid Date",
            "Invalid DateTime",
        ],
    )
    def test_prompt_with_parser_fails(self, answer: str, parser: Parser):
        with replace_stdin(answer) as _, pytest.raises(MaxAttemptsException):
            clypi.prompt("Some prompt", parser=parser, max_attempts=1)

    def test_prompt_with_good_parser(self):
        with replace_stdin("2") as _:
            res = clypi.prompt("Some prompt", parser=lambda x: int(x) * 2)
            assert res == 4

    def test_prompt_with_bad_validate(self):
        with replace_stdin("2") as _, pytest.raises(MaxAttemptsException):
            clypi.prompt("Some prompt", parser=_raise_error, max_attempts=1)



import importlib.util
import inspect
import re
import typing as t
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from types import NoneType, UnionType

from clypi._cli import type_util

HAS_V6E = importlib.util.find_spec("v6e") is not None


class UnparseableException(Exception):
    pass


def dash_to_snake(s: str) -> str:
    return re.sub(r"^-+", "", s).replace("-", "_")


def snake_to_dash(s: str) -> str:
    return s.replace("_", "-")


def normalize_args(args: t.Sequence[str]) -> list[str]:
    new_args: list[str] = []
    for a in args:
        if a.startswith("-") and "=" in a:
            new_args.extend(a.split("=", 1))
        else:
            new_args.append(a)
    return new_args


@dataclass
class Arg:
    value: str
    orig: str
    arg_type: t.Literal["long-opt", "short-opt", "pos"]

    def is_pos(self):
        return self.arg_type == "pos"

    def is_long_opt(self):
        return self.arg_type == "long-opt"

    def is_short_opt(self):
        return self.arg_type == "short-opt"

    def is_opt(self):
        return self.is_long_opt() or self.is_short_opt()


def parse_as_attr(arg: str) -> Arg:
    if arg.startswith("--"):
        return Arg(value=dash_to_snake(arg), orig=arg, arg_type="long-opt")

    if arg.startswith("-"):
        return Arg(value=dash_to_snake(arg), orig=arg, arg_type="short-opt")

    return Arg(value=arg, orig=arg, arg_type="pos")


def _parse_builtin(builtin: type) -> t.Callable[[t.Any], t.Any]:
    def inner(value: t.Any):
        if isinstance(value, list):
            value = value[0]
        return builtin(value)

    return inner


def _parse_list(_type: t.Any) -> t.Callable[[t.Any], list]:
    def inner(value: t.Any):
        if not isinstance(value, list):
            raise ValueError(
                f"Don't know how to parse {value!r} as {type_util.type_to_str(_type)}"
            )

        inner_type = _type.__args__[0]
        parser = from_type(inner_type)
        return [parser(x) for x in value]

    return inner


def _parse_tuple(_type: t.Any) -> t.Callable[[t.Any], tuple]:
    def inner(value: t.Any):
        # Tuples of size 1 are passed in as strings
        if isinstance(value, str):
            value = (value,)

        if not isinstance(value, tuple | list):
            raise ValueError(
                f"Don't know how to parse {value!r} as {type_util.type_to_str(_type)}"
            )

        # TODO: can be made more efficient
        inner_types = _type.__args__
        if inner_types[-1] is Ellipsis:
            inner_types = [inner_types[0]] * len(value)

        if len(inner_types) > len(value):
            raise ValueError(
                f"Not enough arguments for type {type_util.type_to_str(_type)} (got {value})"
            )

        ret = []
        for val, inner_type in zip(value, inner_types):
            ret.append(from_type(inner_type)(val))
        return tuple(ret)

    return inner


def _parse_union(_type: UnionType) -> t.Callable[[t.Any], t.Any]:
    def inner(value: t.Any):
        errors = []
        for a in _type.__args__:
            try:
                return from_type(a)(value)
            except (ValueError, TypeError) as e:
                errors.append(e)
            except UnparseableException:
                pass

        if errors:
            raise errors[0]

    return inner


def _parse_literal(_type: t.Any) -> t.Callable[[t.Any], t.Any]:
    def inner(value: t.Any):
        if isinstance(value, list):
            value = value[0]

        for a in _type.__args__:
            if a == value:
                return value

        raise ValueError(
            f"Value {value} is not a valid choice between {type_util.type_to_str(_type)}"
        )

    return inner


def _parse_none(value: t.Any) -> None:
    if value is not None:
        raise ValueError(f"Value {value!r} is not None")
    return None


def _parse_bool(value: t.Any) -> bool:
    if value.lower() in ("y", "yes", "true"):
        return True
    if value.lower() in ("n", "no", "false"):
        return False
    raise ValueError(f"Value {value!r} is not a valid boolean")


def _parse_enum(_type: type[Enum]):
    def inner(value: t.Any) -> Enum:
        if not isinstance(value, str):
            raise ValueError(
                f"Don't know how to parse {value!r} as {type_util.type_to_str(_type)}"
            )

        for enum_val in _type:
            if value.lower() == enum_val.name.lower():
                return enum_val

        raise ValueError(
            f"Value {value} is not a valid choice between {type_util.type_to_str(_type)}"
        )

    return inner


def from_v6e(_type: t.Any) -> t.Callable[[t.Any], t.Any] | None:
    import v6e as v  # type: ignore

    v6e_builtins: dict[t.Any, t.Callable] = {
        bool: v.bool(),
        int: v.int(),
        float: v.float(),
        str: v.str(),
        Path: v.path(),
        datetime: v.datetime(),
        timedelta: v.timedelta(),
    }
    if _type in v6e_builtins:
        return v6e_builtins[_type]

    return None


@type_util.ignore_annotated
def from_type(_type: t.Any) -> t.Callable[[t.Any], t.Any]:
    if HAS_V6E and (parser := from_v6e(_type)):
        return parser

    if _type is bool:
        return _parse_bool

    if _type in (int, float, str, Path):
        return _parse_builtin(_type)

    if type_util.is_list(_type):
        return _parse_list(_type)

    if type_util.is_tuple(_type):
        return _parse_tuple(_type)

    if isinstance(_type, UnionType):
        return _parse_union(_type)

    if t.get_origin(_type) == t.Literal:
        return _parse_literal(_type)

    if _type is NoneType:
        return _parse_none

    if inspect.isclass(_type) and issubclass(_type, Enum):
        return _parse_enum(_type)

    raise UnparseableException(
        f"Don't know how to parse as {type_util.type_to_str(_type)} ({type(_type)})"
    )


Nargs: t.TypeAlias = t.Literal["*", "+"] | float


@dataclass
class CurrentCtx:
    name: str = ""
    nargs: Nargs = 0
    max_nargs: Nargs = 0

    _collected: list[str] = field(init=False, default_factory=list)

    def has_more(self) -> bool:
        if isinstance(self.nargs, float | int):
            return self.nargs > 0
        return True

    def needs_more(self) -> bool:
        if isinstance(self.nargs, float | int):
            return self.nargs > 0
        elif self.nargs == "+":
            return True
        return False

    def collect(self, item: str) -> None:
        if isinstance(self.nargs, float | int):
            self.nargs -= 1
        elif self.nargs == "+":
            self.nargs = "*"

        self._collected.append(item)

    @property
    def collected(self) -> str | list[str]:
        if self.max_nargs == 1:
            return self._collected[0]
        return self._collected

    def __bool__(self) -> bool:
        return bool(self.name)



from __future__ import annotations

import typing as t
from getpass import getpass

import clypi
from clypi._cli import parser
from clypi._util import _UNSET, Unset
from clypi.configuration import get_config
from clypi.exceptions import AbortException, MaxAttemptsException

MAX_ATTEMPTS: int = 20


def _error(msg: str):
    clypi.print(msg, fg="red")


def _input(
    prompt: str,
    default: T | Unset = _UNSET,
    hide_input: bool = False,
) -> str | T | Unset:
    fun = getpass if hide_input else input
    styled_prompt = get_config().theme.prompts(prompt)
    res = fun(styled_prompt)
    if res:
        return res
    return default


def _display_default(default: t.Any) -> str:
    if isinstance(default, bool):
        return "Y/n" if default else "y/N"
    return f"{default}"


def _build_prompt(text: str, default: t.Any) -> str:
    prompt = text
    if default is not _UNSET:
        prompt += f" [{_display_default(default)}]"
    prompt += ": "
    return prompt


def confirm(
    text: str,
    *,
    default: bool | Unset = _UNSET,
    max_attempts: int = MAX_ATTEMPTS,
    abort: bool = False,
) -> bool:
    """
    Prompt the user for a yes/no value

    :param text: The prompt text.
    :param default: The default value.
    :param max_attempts: The maximum number of attempts to get a valid value.
    :return: The parsed value.
    """
    parsed_inp = prompt(
        text=text,
        default=default,
        max_attempts=max_attempts,
        parser=parser.from_type(bool),
    )
    if abort and not parsed_inp:
        raise AbortException()
    return parsed_inp


T = t.TypeVar("T")

Parser: t.TypeAlias = t.Callable[[t.Any], T]


@t.overload
def prompt(
    text: str,
    *,
    default: str | Unset = _UNSET,
    hide_input: bool = False,
    max_attempts: int = MAX_ATTEMPTS,
) -> str: ...


@t.overload
def prompt(
    text: str,
    *,
    parser: Parser[T] | type[T],
    default: T | Unset = _UNSET,
    hide_input: bool = False,
    max_attempts: int = MAX_ATTEMPTS,
) -> T: ...


def prompt(
    text: str,
    *,
    parser: Parser[T] | type[T] | type[str] = str,
    default: T | Unset = _UNSET,
    hide_input: bool = False,
    max_attempts: int = MAX_ATTEMPTS,
) -> T:
    """
    Prompt the user for a value.

    :param text: The prompt text.
    :param default: The default value.
    :param parser: The parser function parse the input with.
    :param max_attempts: The maximum number of attempts to get a valid value.
    :return: The parsed value.
    """

    # Build the prompt
    prompt = _build_prompt(text, default)

    # Loop until we get a valid value
    for _ in range(max_attempts):
        inp = _input(prompt, default=default, hide_input=hide_input)
        if inp is _UNSET:
            _error("A value is required.")
            continue

        # User answered the prompt -- Parse
        try:
            if t.TYPE_CHECKING:
                parser = t.cast(Parser[T], parser)
            parsed_inp = parser(inp)
        except (ValueError, TypeError) as e:
            _error(f"Unable to parse {inp!r}, please provide a valid value.\n  ↳  {e}")
            continue

        return parsed_inp

    raise MaxAttemptsException(
        f"Failed to get a valid value after {max_attempts} attempts."
    )



import inspect
import typing as t
from enum import Enum
from types import NoneType, UnionType


def ignore_annotated(fun: t.Callable[[t.Any], t.Any]):
    def inner(_type: t.Any):
        if t.get_origin(_type) == t.Annotated:
            _type = _type.__args__[0]
        return fun(_type)

    return inner


@ignore_annotated
def is_list(_type: t.Any) -> t.TypeGuard[list]:
    return t.get_origin(_type) in (list, t.Sequence)


@ignore_annotated
def is_tuple(_type: t.Any) -> t.TypeGuard[tuple]:
    return t.get_origin(_type) is tuple


@ignore_annotated
def tuple_size(_type: t.Any) -> float:
    args = _type.__args__
    if args[-1] is Ellipsis:
        return float("inf")
    return len(args)


@ignore_annotated
def remove_optionality(_type: t.Any) -> t.Any:
    if not isinstance(_type, UnionType):
        return _type

    new_args = []
    for arg in _type.__args__:
        if arg is not NoneType:
            new_args.append(arg)

    if len(new_args) == 1:
        return new_args[0]

    return t.Union[*new_args]


@ignore_annotated
def type_to_str(_type: t.Any) -> str:
    _map = {
        "bool": "boolean",
        "int": "integer",
        "float": "float",
        "str": "string",
        "Path": "Path",
    }
    if inspect.isclass(_type) and _type.__name__ in _map:
        return _map[_type.__name__]

    if t.get_origin(_type) is t.Literal:
        return "{" + "|".join(type_to_str(tp) for tp in _type.__args__) + "}"

    if inspect.isclass(_type) and issubclass(_type, Enum):
        return "{" + "|".join(tp.name for tp in _type) + "}"

    if isinstance(_type, UnionType):
        return "[" + "|".join(type_to_str(tp) for tp in _type.__args__) + "]"

    if is_tuple(_type):
        args = _type.__args__
        if args[-1] is Ellipsis:
            return "(" + type_to_str(_type.__args__[0]) + ", ...)"
        else:
            inner = ", ".join(type_to_str(tp) for tp in _type.__args__)
            maybe_comma = "," if len(args) == 1 else ""  # E.g.: ("foo",)
            return "(" + inner + maybe_comma + ")"

    if is_list(_type):
        return "list[" + ", ".join(type_to_str(tp) for tp in _type.__args__) + "]"

    return str(_type)



from dataclasses import dataclass, field

from clypi._cli.formatter import ClypiFormatter, Formatter
from clypi.colors import Styler
from clypi.exceptions import (
    ClypiException,
)


@dataclass
class Theme:
    usage: Styler = field(default_factory=lambda: Styler(fg="yellow"))
    prog: Styler = field(default_factory=lambda: Styler(bold=True))
    section_title: Styler = field(default_factory=lambda: Styler())

    # Subcommands
    subcommand: Styler = field(default_factory=lambda: Styler(fg="blue", bold=True))

    # Options
    long_option: Styler = field(default_factory=lambda: Styler(fg="blue", bold=True))
    short_option: Styler = field(default_factory=lambda: Styler(fg="green", bold=True))

    # Positionals
    positional: Styler = field(default_factory=lambda: Styler(fg="blue", bold=True))

    type_str: Styler = field(default_factory=lambda: Styler(fg="yellow", bold=True))
    prompts: Styler = field(default_factory=lambda: Styler(fg="blue", bold=True))


@dataclass
class ClypiConfig:
    theme: Theme = field(default_factory=Theme)
    help_formatter: Formatter = field(default_factory=ClypiFormatter)
    help_on_fail: bool = True
    nice_errors: tuple[type[Exception]] = field(
        default_factory=lambda: (ClypiException,)
    )


_CONFIG = ClypiConfig()


def configure(config: ClypiConfig):
    global _CONFIG
    _CONFIG = config


def get_config() -> ClypiConfig:
    return _CONFIG



import typing as t
from pathlib import Path

import clypi

t.assert_type(clypi.confirm("A"), bool)
t.assert_type(clypi.prompt("A"), str)
t.assert_type(clypi.prompt("A", parser=int), int)
t.assert_type(clypi.prompt("A", parser=lambda x: Path(x)), Path)
t.assert_type(clypi.prompt("A", parser=lambda x: Path(x) if x else None), Path | None)



def distance(this: str, other: str) -> int:
    if not this or not other:
        return max(len(this), len(other))

    n, m = len(this), len(other)
    dist = [[0 for _ in range(m + 1)] for _ in range(n + 1)]

    # Prepopulate first X and Y axis
    for t in range(0, n + 1):
        dist[t][0] = t
    for o in range(0, m + 1):
        dist[0][o] = o

    # Compute actions
    for t in range(n):
        for o in range(m):
            insertion = dist[t][o + 1] + 1
            deletion = dist[t + 1][o] + 1
            substitution = dist[t][o] + (1 if this[t] != other[o] else 0)
            dist[t + 1][o + 1] = min(insertion, deletion, substitution)

    # Get bottom right of computed matrix
    return dist[n][m]



import asyncio
from enum import Enum
from pathlib import Path
from typing import Literal

import v6e as v
from typing_extensions import override

import clypi
from clypi import Command, Positional, arg


def debug(fun):
    """
    Just a utility decorator to display the commands being passed in a somewhat
    nice way
    """

    async def inner(self):
        boxed = clypi.boxed(
            clypi.style(self, bold=True), title="Debug", color="magenta"
        )
        print(boxed, end="\n\n")
        await fun(self)

    return inner


class Env(Enum):
    QA = 1
    PROD = 2


class RunParallel(Command):
    """
    Runs all of the files in parallel
    """

    files: Positional[list[str]]
    exceptions_with_reasons: Path | None = arg(
        default=None,
        parser=v.path().exists(),
    )
    env: Env = arg(...)

    @debug
    async def run(self):
        clypi.print(f"{self.env.name} - Running all files", fg="blue", bold=True)

        async with clypi.Spinner(f"Running {', '.join(self.files)} in parallel"):
            await asyncio.sleep(2)

        async with clypi.Spinner(f"Linting {', '.join(self.files)} in parallel"):
            await asyncio.sleep(2)

        clypi.print("\nDone!", fg="green", bold=True)


class RunSerial(Command):
    """
    Runs all of the files one by one
    """

    files: Positional[list[Path]] = arg(parser=v.list(v.path().exists()))
    env: Env = arg(...)

    @debug
    async def run(self):
        clypi.print(f"{self.env.name} - Running all files", fg="blue", bold=True)
        for f in self.files:
            async with clypi.Spinner(f"Running {f.as_posix()} in parallel"):
                await asyncio.sleep(2)
        clypi.print("\nDone!", fg="green", bold=True)


class Run(Command):
    """
    Allows running files with different options
    """

    subcommand: RunParallel | RunSerial
    quiet: bool = False
    env: Env = Env.PROD
    format: Literal["json", "pretty"] = "pretty"


class Lint(Command):
    """
    Lints all of the files in a given directory using the latest
    termuff rules.
    """

    files: Positional[list[str]] = arg(help="The list of files to lint")
    quiet: bool = arg(
        short="q",
        help="If the linter should omit all stdout messages",
        default=False,
    )
    timeout: int = arg(help="Disable the termuff cache")
    index: str = arg(
        default="http://pypi.org",
        help="The index to download termuff from",
        prompt="What index do you want to download termuff from?",
    )

    @debug
    async def run(self):
        async with clypi.Spinner(f"Linting {', '.join(self.files)}"):
            await asyncio.sleep(self.timeout)
        clypi.print("\nDone!", fg="green", bold=True)


class Main(Command):
    """
    Termuff is a powerful command line interfact to lint and
    run arbitrary files.
    """

    subcommand: Run | Lint | None = None
    verbose: bool = arg(short="v", default=False)

    @override
    @classmethod
    def prog(cls):
        return "termuff"

    @override
    @classmethod
    def epilog(cls):
        return "Learn more at http://termuff.org"

    @debug
    async def run(self):
        self.print_help()


if __name__ == "__main__":
    main: Main = Main.parse()
    main.start()
